<?xml version="1.0"?>
<xsl:stylesheet version="1.0"
  xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
  xmlns:axsl="http://www.w3.org/1999/XSL/TransformAlias"
  xmlns:template="http://www.boddie.org.uk/ns/xmltools/template">

  <xsl:output indent="yes"/>
  <xsl:namespace-alias stylesheet-prefix="axsl" result-prefix="xsl"/>



  <!-- Match the document itself. -->

  <xsl:template match="/">
    <axsl:stylesheet version="1.0">

      <axsl:output indent="yes"/>

      <!-- Make document parameters for all elements appearing to use enumerations. -->
      <xsl:for-each select="//element[*/@expr]">
        <axsl:param name="{@name}"/>
      </xsl:for-each>

      <!-- Make a document-level rule. -->
      <axsl:template match="/">
        <axsl:apply-templates select="*"/>
      </axsl:template>

      <!-- Process the top-level element to make other rules. -->
      <xsl:apply-templates select="*"/>

      <!-- Replicate unknown elements. -->
      <axsl:template match="@*|node()">
        <axsl:copy>
          <axsl:apply-templates select="@*|node()"/>
        </axsl:copy>
      </axsl:template>

    </axsl:stylesheet>
  </xsl:template>



  <!-- Match element references. -->

  <xsl:template match="element">
    <xsl:param name="nested">false</xsl:param>

    <!-- Make a rule for the element. -->
    <axsl:template match="{@name}">

      <!-- Copy the element. -->
      <xsl:element name="{@name}">

        <!-- Process attributes. -->
        <axsl:apply-templates select="@*"/>

        <!-- Find elements and determine how to process them. -->
        <xsl:for-each select="element">
          <xsl:choose>
            <!-- Enumerations. -->
            <xsl:when test="@expr">
              <xsl:call-template name="inside-element"/>
            </xsl:when>
            <!-- Top-level or "section" elements are added. -->
            <xsl:when test="count(attribute) = 0 or $nested = 'false'">
              <xsl:element name="{@name}">
                <xsl:call-template name="inside-element"/>
              </xsl:element>
            </xsl:when>
            <!-- Other elements are only added if found. -->
            <xsl:otherwise>
              <axsl:apply-templates select="{@name}"/>
            </xsl:otherwise>
          </xsl:choose>
        </xsl:for-each>

      </xsl:element>

    </axsl:template>

    <!-- Make rules for nested elements. -->
    <xsl:choose>
      <xsl:when test="$nested = 'true'">
        <xsl:apply-templates select="*[not(@expr)]">
          <xsl:with-param name="nested">true</xsl:with-param>
        </xsl:apply-templates>
      </xsl:when>
      <xsl:otherwise>
        <xsl:apply-templates select="*/*[not(@expr)]">
          <xsl:with-param name="nested">true</xsl:with-param>
        </xsl:apply-templates>
      </xsl:otherwise>
    </xsl:choose>

  </xsl:template>



  <!-- Fill in elements. -->

  <xsl:template name="inside-element">
    <axsl:apply-templates select="{@name}/@*"/>
    <xsl:choose>
      <!-- Enumerations. -->
      <xsl:when test="@expr">
        <xsl:call-template name="inside-enumeration"/>
      </xsl:when>
      <xsl:otherwise>
        <!-- Enumeration containers. -->
        <xsl:for-each select="*[@expr]">
          <xsl:call-template name="inside-element"/>
        </xsl:for-each>
        <!-- Normal elements. -->
        <xsl:if test="count(element[not(@expr)]) != 0">
          <axsl:apply-templates select="{@name}/*"/>
        </xsl:if>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>



  <!-- Fill in enumerations. -->

  <xsl:template name="inside-enumeration">
    <axsl:for-each select="${../@name}/{../@name}/{@name}">
      <axsl:copy>
        <axsl:apply-templates select="@*"/>
      </axsl:copy>
    </axsl:for-each>
  </xsl:template>

</xsl:stylesheet>
